# -*- coding: utf-8 -*-
"""
SynthTransactions
A model to create synthetic credit card transaction datasets

v 1.0


Functions:

    NormalizeDistributions    Internal function to ensure distributions are normalised
    
    CreateTransactions        The function to create the data set
                              See below for a list of input parameters
      
                        
Change log:

    v 1.0     Initial version. If you encounter any problem, please contact mzanin@innaxis.org
    
"""


import random
import math



def NormalizeDistributions( D ):
    
    newD = [ value / sum(D) for value in D ]
    return newD
    


'''
Input parameters:

    numUsers            The number of users, i.e. of credit cards to be simulated
    numStores           The number of stores to be simulated, receiving the transactions
    numDays             Number of days to simulate
    variability         Parameters that allows the creation of burst of transactions
    userSwap            Boolean value enabling the swap of users
    userSwapProb        Probability of swapping two users - if the condition about
                           distance is fulfilled
    userSwapDist        Max distance between which two users can be swapped
'''

def CreateTransactions( numUsers, numStores, numDays, variability = 1.1, \
                        userSwap = False, userSwapProb = 0.1, userSwapDist = 10.0 ):

    from SynthTransactions_InputData import HourlyDistr as HD
    from SynthTransactions_InputData import DailyDistr as DD
    from SynthTransactions_InputData import QuantityDistr as QD
    from SynthTransactions_InputData import NumOpsDistr as NOD
    from SynthTransactions_InputData import NumOpsDistrStores as NOD_S

    HD = NormalizeDistributions( HD )
    DD = NormalizeDistributions( DD )
    QD = NormalizeDistributions( QD )
    NOD = NormalizeDistributions( NOD )
    NOD_S = NormalizeDistributions( NOD_S )
    
    
    Transactions = []    
    UsersAvgFreq = []
    UsersInhib = []

    
    for user in range( numUsers ):
        freqP = random.random()
        freq = 1
        for k in range( 100 ):
            freqP -= NOD[k]
            if freqP < 0.0:
                break
            freq += 1
        UsersAvgFreq.append( freq )
        
    for user in range( numUsers ):
        UsersInhib.append( 0.0 )
        
    
    StoresW = []
    for store in range( numStores ):
        freqP = random.random()
        freq = 1
        for k in range( 50 ):
            freqP -= NOD_S[k]
            if freqP < 0.0:
                break
            freq += 1
        StoresW.append( freq * 20 )
    
    StoresW = [ float(i) / sum( StoresW ) for i in StoresW ]
    
        
        
    for day in range( numDays ):
        
        dayOfTheWeek = day % 7
        
        for user in range( numUsers ):

            numOpsDay = ( UsersAvgFreq[ user ] / 30.0 ) * ( DD[ dayOfTheWeek ] * 7.0 )
            numOpsDay -= UsersInhib[ user ]
            numOpsDay = random.gauss( numOpsDay, variability )
            
            if numOpsDay < 0.0:
                UsersInhib[ user ] = - numOpsDay
                numOpsDay = 0.0
                continue
            else:
                UsersInhib[ user ] = 0.0
            
            a = numOpsDay - math.floor( numOpsDay )
            if random.random() < a:
                numOpsDay = math.ceil( numOpsDay )
            else:
                numOpsDay = math.floor( numOpsDay )
            
            for trans in range( int( round(numOpsDay) ) ):
                
                hourP = random.random()
                hour = 0
                for k in range( 24 ):
                    hourP -= HD[k]
                    if hourP < 0.0:
                        break
                    hour += 1
                
                quantP = random.random()
                quant = 25.0 / 2.0
                for k in range( 24 ):
                    quantP -= QD[k]
                    if quantP < 0.0:
                        break
                    quant += 25.0
                
                storeP = random.random()
                store = 0
                for k in range( len(StoresW) ):
                    storeP -= StoresW[k]
                    if storeP < 0.0:
                        break
                    store += 1
                if store >= len(StoresW):
                    store = len(StoresW)
                    
                Transactions.append( ( day, user, hour, quant, store ) )
            
        if userSwap:
            for user1 in range( numUsers ):
                for user2 in range( numUsers ):
                    
                    if user1 == user2:
                        continue
                    
                    if abs( UsersAvgFreq[ user1 ] - UsersAvgFreq[ user2 ] ) >= userSwapDist:
                        continue
                    
                    if random.random() > userSwapProb:
                        continue
                    
                    UsersAvgFreq[ user1 ], UsersAvgFreq[ user2 ] = UsersAvgFreq[ user2 ], UsersAvgFreq[ user1 ]
    
    return Transactions


